
/*
 * Momentum window. May 2017
 * Combines simulated and experimental data.
 */

void MomentumWindow(Double_t xscale = 0.97, Int_t bins = 2000, Double_t emin = 0, Double_t emax = 2000, Bool_t vetoOn = false) {

    TString dir = "../../output/scan/";
    TString expdir = "/home/oliskir/Desktop/F20-IGISOL/data/sorted/";
    TString ss = "_c";
    
    double calib = 147.2; // Bmax (T) => I/Imax (%)
    calib *= xscale;
    
    const int ncuts = 3;
    double cuts[] = {10, 1200, 1670}; // keV

    // simulated data
    vector<double> bmax = {0.110, 0.115, 0.120, 0.125, 0.130, 0.135, 0.140, 0.145, 0.155, 0.160, 0.165, 0.170, 0.175};
    
    vector<TString> file;
    for (auto& b : bmax) {
        TString fname = dir + Form("bmax%.0f.root", b*1E4); 
        file.push_back(fname);
    }

    for (auto&b : bmax) b *= calib;    

    // source activity
    double A0    = 398e3; // Start activity in Bq
    double years = 8.0;   // years since source was made
    double t12   = 31.6;  // half-life
    double A     = A0 * TMath::Exp(- years * TMath::Log(2) / t12);
    cout << "Activity: " << A << " Bq" << endl;
    
    // intensity of 1.6-MeV line
    double intensity = (0.2522 + 0.0403)/1E3;

    // duration of measurement in minutes
    double duration = 1;
        
    // get histograms
    vector<TH1F*> hist;
    Int_t i=0;
    for (auto& fname : file) {
        TFile * f = new TFile(fname, "READ");
        TString hname = Form("sim%i", i+1);
        hist.push_back(new TH1F(hname, hname, bins, emin, emax));
        TTree * t = (TTree*)f->Get("Detector");

        // Normalization
        Double_t norm = ((TParameter<double>*)(t->GetUserInfo()->FindObject("NORMALIZATION")))->GetVal();
	    // Calculate scaling factor
	    Double_t sf = intensity * 60 * duration * A / norm;
        if (i==0) cout << "Scaling factor: " << sf << endl;        

        // veto cut
        TCut cut("");
        if (vetoOn) cut = TCut("EDepVeto == 0");
        
        // draw
        t->Draw("EDepSignal>>"+hname, cut);
        
        // scale
        hist[i]->Scale(sf);
        
        i++;

        cout << fname << " " << hname << endl;
    }

    // do integrals      
    vector<double> integrals[ncuts];
    cout << " I/I_max (%)    Counts/min    Peak frac. (%)" << endl;
    i=0;
    for (auto& h : hist) {
    
        vector<double> counts;
        for (int j = 0; j < ncuts; j++) {
            counts.push_back(h->Integral(h->FindBin(cuts[j]), h->FindBin(emax)));
            integrals[j].push_back(counts[j]);
        }
        
        double frac = counts[2] / counts[0];
    
        cout << Form("  %.1f", bmax[i]) 
             << Form("    %.0f", counts[0])
             << Form("    %.0f%%", frac*100.) << endl;


        i++;
    }


    //----------------------------------------------------------------------//
    const int expncuts = 2;
    double expcuts[] = {1200, 1550}; // keV

    // experimental data
    vector<TString> expfile;
    vector<double> current;
    expfile.push_back(expdir+"s0169_00.root"); current.push_back(28.0);
    expfile.push_back(expdir+"s0170_00.root"); current.push_back(26.0);
    expfile.push_back(expdir+"s0171_00.root"); current.push_back(24.0);
    expfile.push_back(expdir+"s0181_00.root"); current.push_back(23.0);
    expfile.push_back(expdir+"s0172_00.root"); current.push_back(22.0);
    expfile.push_back(expdir+"s0176_00.root"); current.push_back(21.0);
    expfile.push_back(expdir+"s0173_00.root"); current.push_back(20.0);
    expfile.push_back(expdir+"s0182_00.root"); current.push_back(19.0);
    expfile.push_back(expdir+"s0174_00.root"); current.push_back(18.0);
    expfile.push_back(expdir+"s0175_00.root"); current.push_back(16.0);
    
    double expduration = 5;

    // get histograms
    vector<TH1F*> exphist;
    i=0;
    for (auto& fname : expfile) {
        TFile * f = new TFile(fname, "READ");
        TString hname = Form("exp%i", i+1);
        exphist.push_back(new TH1F(hname, hname, bins, emin, emax));
        TTree * t = (TTree*)f->Get("data");
        TCut cut("");
        if (vetoOn) cut = TCut("M1<=1 && E1<140+0.21*E0");
        t->Draw("E0>>"+hname, cut);
        exphist[i]->Scale(1./expduration);
        i++;
        cout << fname << " " << hname << endl;
    }    

    // do exp integrals      
    vector<double> expintegrals[expncuts];
    cout << " I/I_max (%)    Counts/min" << endl;
    i=0;
    for (auto& h : exphist) {
    
        vector<double> counts;
        for (int j = 0; j < expncuts; j++) {
            counts.push_back(h->Integral(h->FindBin(expcuts[j]), h->FindBin(emax)));
            expintegrals[j].push_back(counts[j]);
        }
    
        cout << Form("  %.0f%%", current[i]) 
             << Form("    %.0f", counts[0]) << endl;


        i++;
    }
    //----------------------------------------------------------------------//

    // canvas
    TCanvas * c1 = new TCanvas("c1");

	// Legend
	Double_t xl1=.63, yl1=0.5, xl2=xl1+0.22, yl2=yl1+0.35;
    TLegend *leg = new TLegend(xl1, yl1, xl2, yl2);

    // TGraphs
    const int M = ncuts;
    TGraph* gr[M];
    for (int i = 0; i < M; i++) {
        gr[i] = new TGraph(integrals[i].size(), bmax.data(), integrals[i].data());      
    }
    
    // Multigraph
    TMultiGraph *mg = new TMultiGraph();
    for (int j = 0; j < M; j++) {
        gr[j]->SetLineColor(j+1);
        gr[j]->SetLineWidth(2);
        mg->Add(gr[j]);
        leg->AddEntry(gr[j], Form("SIM [E > %.0f keV}", cuts[j]), "l");
    }

    mg->SetTitle("Momentum window (1.69 MeV)");
    mg->Draw("AC");

    // experimental data    
    const int N = expncuts;
    TGraphErrors * expgr[N];
    for (int i = 0; i < N; i++) {
        vector<double> ex, ey;
        for (auto& y : expintegrals[i]) {
            ex.push_back(0);
            ey.push_back(sqrt(y));
        }
        expgr[i] = new TGraphErrors(expintegrals[i].size(), current.data(), expintegrals[i].data(), ex.data(), ey.data());      

        int col = 4;
        if (i == 1) col = 6;
        expgr[i]->SetLineColor(col);
        expgr[i]->SetLineStyle(1+i);
        expgr[i]->SetMarkerColor(col);
        expgr[i]->SetMarkerStyle(20+i);
        expgr[i]->Draw("samePL");

        leg->AddEntry(expgr[i], Form("EXP [E > %.0f keV]", expcuts[i]), "p");
    }



    auto yaxis = mg->GetYaxis();
    yaxis->SetTitle("Counts/minute");
    auto xaxis = mg->GetXaxis();
    xaxis->SetTitle("I/I_{max} (%)");
    xaxis->SetLimits(14, 30);             // along X
    mg->GetHistogram()->SetMaximum(1500); // along          
    mg->GetHistogram()->SetMinimum(0);  //   Y     

    leg->Draw("Same");

    // text
   TText *t = new TText(14.5, 1300, Form("x-scaling=%.3f", scale));
   t->SetTextAlign(11);
   t->SetTextColor(1);
   t->SetTextFont(43);
   t->SetTextSize(20);
   t->Draw("same");

    // make pretty    
    gPad->SetLogy(0);
    gPad->SetGrid(0,1);
    gPad->SetTicks();
    
    gStyle->SetOptStat("");
    
    TString fname = "/home/oliskir/Dropbox/Work_ln-s/F20-IGISOL/analysis/figures/";
    fname += "momentum_window_sim" + ss + ".png";
    c1->Print(fname);                  
}
