//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//

#include "VKEventAction.hh"
#include "VKScintillatorHit.hh"
#include "VKSiPMHit.hh"
#include "VKUserEventInformation.hh"
#include "VKTrajectory.hh"
#include "VKRecorderBase.hh"

#include "VKScintillatorSensitiveDetector.hh"

#include "G4EventManager.hh"
#include "G4SDManager.hh"
#include "G4RunManager.hh"
#include "G4Event.hh"
#include "G4EventManager.hh"
#include "G4TrajectoryContainer.hh"
#include "G4Trajectory.hh"
#include "G4VVisManager.hh"
#include "G4ios.hh"
#include "G4UImanager.hh"
#include "G4SystemOfUnits.hh"

#include <vector>

using namespace std;


VKEventAction::VKEventAction(VKRecorderBase* r)
 : fRecorder(r),
   fSaveThreshold(0),
   fScintCollID(-1),
   fVetoCollID(-1),
   fSiPMCollID(-1),
   fVerbose(0),
   fSiPMThreshold(1),
   fForcedrawphotons(false),
   fForcenophotons(false),
   fEvtPct(0)
{
    fEventMessenger = new VKEventMessenger(this);
}


VKEventAction::~VKEventAction()
{
    // do nothing
}


void VKEventAction::BeginOfEventAction(const G4Event* anEvent)
{
    // New event, add the user information object
    G4EventManager::GetEventManager() -> SetUserInformation(new VKUserEventInformation);

    G4SDManager* SDman = G4SDManager::GetSDMpointer();
    if (fScintCollID < 0) fScintCollID = SDman -> GetCollectionID("Signal_SD_HitCollection");
    if (fVetoCollID < 0)  fVetoCollID  = SDman -> GetCollectionID("Veto_SD_HitCollection");
    if (fSiPMCollID < 0)  fSiPMCollID  = SDman -> GetCollectionID("SiPM_SD_HitCollection");
    
    if (fRecorder) fRecorder -> RecordBeginOfEvent(anEvent);
}


void VKEventAction::EndOfEventAction(const G4Event* anEvent)
{
    // event info
    VKUserEventInformation* eventInfo
        = (VKUserEventInformation*)anEvent -> GetUserInformation();

    // trajectories 
    G4TrajectoryContainer* trajectoryContainer = anEvent -> GetTrajectoryContainer(); 

    // no. of trajectories
    G4int n_trajectories = 0;
    if (trajectoryContainer) n_trajectories = trajectoryContainer -> entries();

    // extract the trajectories and draw them
    if (G4VVisManager::GetConcreteInstance()) {
        for (G4int i=0; i<n_trajectories; i++) {
            VKTrajectory* trj = (VKTrajectory*)((*(trajectoryContainer))[i]);
            if (trj->GetParticleName() == "opticalphoton") {
                trj -> SetForceDrawTrajectory(fForcedrawphotons);
                trj -> SetForceNoDrawTrajectory(fForcenophotons);
            }
        trj -> DrawTrajectory();
        }
    }
 
    VKScintillatorHitsCollection * scintHC = 0;
    VKScintillatorHitsCollection * vetoHC  = 0;
    VKSiPMHitsCollection* sipmHC = 0;
    G4HCofThisEvent* hitsCE = anEvent -> GetHCofThisEvent();
 
    // Get hit collections
    if (hitsCE) {
        if (fScintCollID >= 0) scintHC = (VKScintillatorHitsCollection*)(hitsCE->GetHC(fScintCollID));
        if (fVetoCollID >= 0)   vetoHC = (VKScintillatorHitsCollection*)(hitsCE->GetHC(fVetoCollID));
        if (fSiPMCollID >= 0)   sipmHC = (VKSiPMHitsCollection*)(hitsCE->GetHC(fSiPMCollID));
    }

    // Hits in scintillators (signal and veto)
    vector<VKScintillatorHitsCollection*> scintHCs = {scintHC, vetoHC};
    G4int id = 0;
    for (auto& hc : scintHCs) {
        if (hc) {
            int n_hit = hc -> entries(); // number of hits
            G4ThreeVector eWeightPos(0.); // energy-weighted position of hit
            G4double edep;
            G4double edepMax = 0;

            //gather info on hits in scintillator
            for (int i=0; i<n_hit; i++) { 
                edep = (*hc)[i] -> GetEdep();
                eventInfo -> IncEDep(id, edep); // add up the energy deposition
                eWeightPos += (*hc)[i]->GetPos() * edep; // calculate energy-weighted position
                if (edep > edepMax) {
                    edepMax = edep; // store max energy deposition
                    G4ThreeVector posMax = (*hc)[i]->GetPos();
                    eventInfo -> SetPosMax(id, posMax, edep);
                }
            }
            if (eventInfo->GetEDep(id) == 0.) {
                if (fVerbose>0) G4cout << "No hits in " << hc->GetName() << " this event." << G4endl;
            }
            else{
                // Finish calculation of energy-weighted position
                eWeightPos /= eventInfo->GetEDep(id);
                eventInfo -> SetEWeightPos(id, eWeightPos);
                if (fVerbose > 0) {
                    G4cout << "\tEnergy weighted position of hits in " << hc->GetName() << " : "
                       << eWeightPos/mm << G4endl;
                }
            }
            
            // print total energy deposition
            if (fVerbose > 0) {
                G4cout << "\tTotal energy deposition in " << hc->GetName() << " : "
                       << eventInfo->GetEDep(id) / keV << " (keV)" << G4endl;
            }
        }
        id++;
    }
    
    G4String scintName[2] = {"Signal", "Veto"};
    
    // Hits in SiPM
    if (sipmHC) {
        G4ThreeVector reconPos(0.,0.,0.);
        G4int sipms = sipmHC->entries();  // number of hits
        
        //Gather info from all SiPMs
        for (G4int i=0; i<sipms; i++) {
            G4int id = (*sipmHC)[i]->GetSiPMNumber();
            G4String name = (id < 2)? scintName[id] : "Unknown";
            if(fVerbose > 0) {
                G4cout << "\t======= SiPM data for: " << name << " =======" << G4endl;
            }
            
            G4int photons = (*sipmHC)[i]->GetPhotonCount();
            eventInfo -> IncHitCount(id, photons);
            reconPos += (*sipmHC)[i]->GetSiPMPos() * photons;
            if (photons >= fSiPMThreshold) {
                eventInfo -> IncSiPMsAboveThreshold(id);
            }
            else { // if it wasnt above the threshold, turn it back off
                (*sipmHC)[i]->SetDrawit(false);
            }

            // if there were any hits, determine position
            if (eventInfo->GetHitCount(id) > 0) {
                reconPos /= eventInfo->GetHitCount(id);
                if(fVerbose > 0) {
                    G4cout << "\tReconstructed position of hits in SiPMs: "
                           << reconPos/mm << G4endl;
                }
                eventInfo -> SetReconPos(id, reconPos);
            }

            //End of event output. later to be controlled by a verbose level
            if (fVerbose > 0) {
                G4cout << "\tNumber of photons that hit SiPMs in this event : "
                       << eventInfo->GetHitCount(id) << G4endl;
                G4cout << "\tNumber of SiPMs above threshold ("<<fSiPMThreshold<<") : "
                       << eventInfo->GetSiPMsAboveThreshold(id) << G4endl;
                G4cout << "\tNumber of photons produced by scintillation in this event : "
                       << eventInfo->GetPhotonCount_Scint(id) << G4endl;
                G4cout << "\tNumber of photons produced by cerenkov in this event : "
                       << eventInfo->GetPhotonCount_Ceren(id) << G4endl;
                G4cout << "\tNumber of photons absorbed (OpAbsorption) in this event : "
                       << eventInfo->GetAbsorptionCount(id) << G4endl;
                G4cout << "\tNumber of photons absorbed at boundaries (OpBoundary) in "
                       << "this event : " << eventInfo->GetBoundaryAbsorptionCount(id)
                       << G4endl;
                G4cout << "Unacounted for photons in this event : "
                       << (eventInfo->GetPhotonCount_Scint(id) +
                           eventInfo->GetPhotonCount_Ceren(id) -
                           eventInfo->GetAbsorptionCount(id) -
                           eventInfo->GetHitCount(id) -
                           eventInfo->GetBoundaryAbsorptionCount(id))
                       << G4endl;
            }
        }

        sipmHC -> DrawAllHits();
    }
  
    // If we have set the flag to save 'special' events, save here
    if (fSaveThreshold && eventInfo->GetPhotonCount() <= fSaveThreshold)
        G4RunManager::GetRunManager()->rndmSaveThisEvent();

    // Print progress
    PrintProgress();

    if (fRecorder) fRecorder -> RecordEndOfEvent(anEvent);
}


/*
 * Sets the save threshold for the random number seed. If the number of photons
 * generated in an event is lower than this, then save the seed for this event
 * in a file called run###evt###.rndm
 */
void VKEventAction::SetSaveThreshold(G4int save)
{
    fSaveThreshold = save;
    G4RunManager::GetRunManager() -> SetRandomNumberStore(true);
    G4RunManager::GetRunManager() -> SetRandomNumberStoreDir("random/");
    //  G4UImanager::GetUIpointer()->ApplyCommand("/random/setSavingFlag 1");
}


void VKEventAction::PrintProgress() 
{
    // number of events to be processed
    double evtMax = G4RunManager::GetRunManager()
                        ->GetCurrentRun()->GetNumberOfEventToBeProcessed();
    // current event no.                        
    double evt = G4RunManager::GetRunManager()
                        ->GetCurrentEvent()->GetEventID();
                        
    G4int evtPct = (evt + 0.5) / evtMax * 100;
    if (evtPct > fEvtPct && evt > 0)
    {
        fEvtPct = evtPct;
        printf(" %i%% of %i events simulated ...", evtPct, (int)evtMax); 
        fflush(stdout); 
        printf(" \r");
    }
}
