//
// ********************************************************************
// * License and Disclaimer                                           *
// *                                                                  *
// * The  Geant4 software  is  copyright of the Copyright Holders  of *
// * the Geant4 Collaboration.  It is provided  under  the terms  and *
// * conditions of the Geant4 Software License,  included in the file *
// * LICENSE and available at  http://cern.ch/geant4/license .  These *
// * include a list of copyright holders.                             *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.  Please see the license in the file  LICENSE  and URL above *
// * for the full disclaimer and the limitation of liability.         *
// *                                                                  *
// * This  code  implementation is the result of  the  scientific and *
// * technical work of the GEANT4 collaboration.                      *
// * By using,  copying,  modifying or  distributing the software (or *
// * any work based  on the software)  you  agree  to acknowledge its *
// * use  in  resulting  scientific  publications,  and indicate your *
// * acceptance of all terms of the Geant4 Software license.          *
// ********************************************************************
//

#include "VKPrimaryGeneratorAction.hh"
#include "VKPrimaryGeneratorMessenger.hh"

#include "G4Event.hh"
#include "G4ParticleTable.hh"
#include "G4ParticleDefinition.hh"
#include "G4SystemOfUnits.hh"
#include "globals.hh"
#include "G4GeneralParticleSource.hh"
#include "G4Electron.hh"

#include <TF1.h>
#include <TMath.h>


VKPrimaryGeneratorAction::VKPrimaryGeneratorAction()
 : fCRMuonAngDistOn(false),
   fBetaSpectrumOn(false),
   fDiskShapedSourceOn(false),
   fGPSourceOn(false),
   fPower(1.85),
   fCRMuonCoverage(1),
   fZenithMin(0*deg),
   fZenithMax(90*deg),
   fSourceInnerDiameter(0*mm),
   fSourceOuterDiameter(7*mm),
   fSourceNormal(G4ThreeVector(0,0,1)),
   fElectronMass(G4Electron::ElectronDefinition()->GetPDGMass()),
   fW0(14.70), // equivalent to E0 = 7.0 MeV
   fWMin(0.),
   fWMax(fW0),
   fPosCentre(G4ThreeVector(0,0,0))   
{
    // create messenger 
    primGen_msg = new VKPrimaryGeneratorMessenger(this);

    // General particle source
    fGPSource = new G4GeneralParticleSource();
    fGPSource -> SetCurrentSourceIntensity(1);
  
    // Table of particles
    G4ParticleTable* particleTable = G4ParticleTable::GetParticleTable();
    G4ParticleDefinition * electron = particleTable -> FindParticle("e-");

    // electron is default particle 
    fGPSource -> GetCurrentSource() -> SetParticleDefinition(electron);

    // Default energy, position, momentum
    fGPSource -> GetCurrentSource() -> GetPosDist() 
              -> SetCentreCoords(G4ThreeVector(0.0 , 0.0, -30.0*cm));
    fGPSource -> GetCurrentSource() -> GetAngDist() 
              -> SetParticleMomentumDirection(G4ThreeVector(0., 0., 1.));
    fGPSource -> GetCurrentSource() -> GetEneDist() 
              -> SetMonoEnergy(1.*MeV);          

    // Initialize radial distribution (for sampling initial position)
    InitRadialDist();
    
    // Get orthogonal vectors perpendicular to fSourceNormal
    GetOrthogonalVectors(fSourceNormal, fOrtho1, fOrtho2);
}


VKPrimaryGeneratorAction::~VKPrimaryGeneratorAction()
{
    delete fGPSource;
    delete fCRMuonZenithDist;
    delete fCRMuonAzimuthalDist;
    delete fRadialDist;
}


void VKPrimaryGeneratorAction::GeneratePrimaries(G4Event* anEvent)
{
    // Is this the first event of the run?
    if (anEvent -> GetEventID() == 0) {
        // if so, store centre position
        fPosCentre = fGPSource -> GetCurrentSource() -> GetPosDist() -> GetCentreCoords(); 
    }

    // Sample (relative) position coordinates if needed
    if (fDiskShapedSourceOn) 
    {
        // Centre position
        
        // polar coordinates in plane
        G4double angle  = G4UniformRand() * 2*CLHEP::pi *rad;
        G4double radius = fRadialDist -> GetRandom() *cm;
        // calculate position
        fPos = fPosCentre + radius*sin(angle)*fOrtho1 + radius*cos(angle)*fOrtho2;
        fGPSource -> GetCurrentSource() -> GetPosDist() -> SetCentreCoords(fPos);
    }

    // Cosmic-ray muon angular distributions
    if (fCRMuonAngDistOn) 
    {
        // sample zenith angle
        double zen = fCRMuonZenithDist -> GetRandom(fZenithMin/rad, fZenithMax/rad);

        // sample azimuthal angle
        double phi = fCRMuonAzimuthalDist -> GetRandom();
        
        // set direction
        double cz = cos(zen);
        double sz = sin(zen);
        double cp = cos(phi);
        double sp = sin(phi);
        fDirection = G4ThreeVector(sz*cp, -cz, sz*sp);
        fGPSource -> GetCurrentSource() -> GetAngDist() 
                                                -> SetParticleMomentumDirection(fDirection);
    }
    
    // Beta decay energy spectrum
    if (fBetaSpectrumOn) 
    {
        // sample energy
        G4double ekin = (fBetaSpectrum->GetRandom(fWMin, fWMax) - 1.) * fElectronMass;
        fGPSource -> GetCurrentSource() -> GetEneDist() -> SetMonoEnergy(ekin);   
    }
    
    // generate primary
    fGPSource -> GeneratePrimaryVertex(anEvent);
}


void VKPrimaryGeneratorAction::InitCRMuonAngDist()
{
    // functional form
    TString function = Form("sin(x)*pow(cos(x), %f)", fPower);
    // print info
//    G4cout.precision(1);
    G4cout << " ---> Initializing angular distributions for CR muons:" << G4endl;
    G4cout << " ---> f(x) = " << function;
    G4cout << ",  " << fZenithMin/deg << " deg <= x <= " << fZenithMax/deg << " deg" << G4endl;
    // initialize TF1s
    fCRMuonZenithDist    = new TF1("zenithDist", function, 0., CLHEP::pi/2.);
    fCRMuonAzimuthalDist = new TF1("azimuthalDist", "1",   0., 2*CLHEP::pi);
}


void VKPrimaryGeneratorAction::InitRadialDist()
{
    double rmin = 0.5 * fSourceInnerDiameter /cm;
    double rmax = 0.5 * fSourceOuterDiameter /cm;
    fRadialDist = new TF1("radialDist", "x", rmin, rmax);
}



void VKPrimaryGeneratorAction::SetCRMuonAngDistPower(float power) 
{
    fPower = power;
    InitCRMuonAngDist();
}


void VKPrimaryGeneratorAction::SetCRMuonAngDistOn(G4bool b) 
{ 
    fCRMuonAngDistOn = b; 
    if (fCRMuonAngDistOn) InitCRMuonAngDist();
}


void VKPrimaryGeneratorAction::SetCRMuonMinZenithAngle(G4double angle) 
{
    fZenithMin = angle;
    InitCRMuonAngDist();
}


void VKPrimaryGeneratorAction::SetCRMuonMaxZenithAngle(G4double angle) 
{
    fZenithMax = angle;
    InitCRMuonAngDist();
}


void VKPrimaryGeneratorAction::GetOrthogonalVectors(const G4ThreeVector& n, G4ThreeVector& o1, G4ThreeVector& o2)
{
    o1 = n.cross(G4ThreeVector(1,0,0));
    if (o1.mag()==0) o1 = n.cross(G4ThreeVector(0,1,0));
    o1 = o1.unit();
    o2 = n.cross(o1);
}


void VKPrimaryGeneratorAction::SetDiskShapedSourceOn(G4bool b) 
{ 
    fDiskShapedSourceOn = b; 
}


void VKPrimaryGeneratorAction::SetSourceInnerDiameter(G4double d)
{
    G4cout << " ---> Source inner diameter set to " << d/mm << " mm" << G4endl;
    fSourceInnerDiameter = d;
    InitRadialDist();
}


void VKPrimaryGeneratorAction::SetSourceOuterDiameter(G4double d)
{
    G4cout << " ---> Source outer diameter set to " << d/mm << " mm" << G4endl;
    fSourceOuterDiameter = d;
    InitRadialDist();
}


void VKPrimaryGeneratorAction::SetSourceNormal(const G4ThreeVector& n)
{
    fSourceNormal = n;
    GetOrthogonalVectors(fSourceNormal, fOrtho1, fOrtho2);
}


// beta spectrum allowed shape (with F=1)
// x is the total beta energy in units of m_e
void VKPrimaryGeneratorAction::InitBetaSpectrum()
{
    fBetaSpectrum = new TF1("betaSpectrum", "x*pow(x*x-1.,0.5)*pow([0]-x,2)", 1., fW0);
    fBetaSpectrum -> SetParameter(0, fW0);
    // print info
//    G4cout.precision(3);
    G4cout << " ---> Initializing beta-decay energy spectrum (allowed shape) with";
    G4cout << " end-point energy of " << (fW0-1.)*fElectronMass/MeV << " MeV" << G4endl;
}


void VKPrimaryGeneratorAction::SetBetaSpectrumOn(G4bool b) 
{ 
    fBetaSpectrumOn = b; 
    if (fBetaSpectrumOn) InitBetaSpectrum();
}


void VKPrimaryGeneratorAction::SetBetaEndPointEnergy(G4double endPoint) 
{ 
    fW0 = (endPoint + fElectronMass)/fElectronMass;
    InitBetaSpectrum();
    // if max sampling energy exceeds new end-point energy, reduce it
    double wMax = fWMax;
    SetBetaMaxEnergy(endPoint);
    fWMax = TMath::Min(wMax, fWMax);
}

void VKPrimaryGeneratorAction::SetBetaMinEnergy(G4double emin)      
{ 
    fWMin = (emin + fElectronMass)/fElectronMass;
}
        
void VKPrimaryGeneratorAction::SetBetaMaxEnergy(G4double emax)      
{ 
    fWMax = (emax + fElectronMass)/fElectronMass;     
}


double VKPrimaryGeneratorAction::GetBetaCoverage() const
{
    double coverage = 1;
    double fullIntegral  = fBetaSpectrum -> Integral(1., fW0);
    double restrIntegral = fBetaSpectrum -> Integral(fWMin, fWMax);
    if (fullIntegral > 0) coverage *= restrIntegral / fullIntegral;
    return coverage;
}


double VKPrimaryGeneratorAction::GetCRMuonCoverage() const
{
    double coverage = 1;
    double fullIntegral  = fCRMuonZenithDist -> Integral(0., CLHEP::pi/2.);
    double restrIntegral = fCRMuonZenithDist -> Integral(fZenithMin/rad, fZenithMax/rad); 
    if (fullIntegral > 0) coverage = restrIntegral / fullIntegral;   
    return coverage;
}


double VKPrimaryGeneratorAction::GetCoverage() const
{
    double coverage = 1;

    if (fBetaSpectrumOn) coverage *= GetBetaCoverage();
    
    if (fCRMuonAngDistOn) {
        coverage *= GetCRMuonCoverage();    
    }
    else {
        // theta distribution
        G4double thMin = fGPSource -> GetCurrentSource() -> GetAngDist() -> GetMinTheta();
        G4double thMax = fGPSource -> GetCurrentSource() -> GetAngDist() -> GetMaxTheta();
        coverage *= (thMax - thMin) / (180. * deg);

        // phi distribution
        G4double phiMin = fGPSource -> GetCurrentSource() -> GetAngDist() -> GetMinPhi();
        G4double phiMax = fGPSource -> GetCurrentSource() -> GetAngDist() -> GetMaxPhi();
        coverage *= (phiMax - phiMin) / (360. * deg);
    }

    return coverage;
}


G4String VKPrimaryGeneratorAction::GetParticleName() const 
{
    return fGPSource -> GetCurrentSource() -> GetParticleDefinition() -> GetParticleName();
}


G4double VKPrimaryGeneratorAction::GetSourceArea() const 
{
    G4double A = 0;
    if (fDiskShapedSourceOn) {
        G4double ro = fSourceOuterDiameter/2;
        G4double ri = fSourceInnerDiameter/2;
        A = CLHEP::pi * (pow(ro, 2) - pow(ri, 2));
    }
    return A;
}

